structure DeflateHuffman :> DeflateHuffman =
struct

	datatype tree = none | leaf of int | node of tree * tree

	exception Insert_at_used_location
	exception No_step
	
	fun insert (none) [] code = leaf code
	  | insert (node (zero,one)) (false::es) code = 
	  		node (insert zero es code, one)
	  | insert (node (zero,one)) (true::es) code =
  			node (zero, insert one es code)
	  | insert (none) (false::es) code =
  			node (insert none es code, none)
	  | insert (none) (true::es) code =
  			node (none, insert none es code)
	  | insert _ es code = raise Insert_at_used_location

	fun at (node (zero, _)) (false::es) = at zero es
	  | at (node (_, one)) (true::es) = at one es
	  | at tree [] = (tree, [])
	  | at _ es = (none, es)

	fun code tree es = 
		let 
			fun impl (leaf c, _) = SOME c
			  | impl _ = NONE 
		in 
			impl (at tree es)
		end

	fun has_step (node _) = true
	  | has_step _ = false
	  
	fun step (node (zero,_)) false = zero
	  | step (node (_, one)) true = one
	  | step _ _ = raise No_step
	   
	fun has_length _ 0 = true
	  | has_length (node (zero, one)) len = has_length one (len - 1) orelse has_length zero (len - 1)
	  | has_length _ _ = false
	  		
	fun collapse tree =
		let
			fun sorted ((_, c1),(_, c2)) = c1 <= c2
			fun impl prefix (leaf c) = [(rev prefix, c)]
			  | impl prefix (none) = []
			  | impl prefix (node (zero, one)) = 
			  		(impl (false::prefix) zero) @ (impl (true::prefix) one)
		in
			DeflateList.sort sorted (impl [] tree)
		end 

	(* for collapse *)
	fun stringify (es, c) = (DeflateBits.to_string es, c)
	
	fun collapse_string tree = map stringify (collapse tree)
	
	(* like collapse_string with ellipsis of successive items *)
	fun collapse_summarised hutree =
		let
			fun succeeding ((p1,c1),(p2,c2)) = length p1 = length p2
				andalso DeflateBits.to_int_msb_first p1 + 1 = DeflateBits.to_int_msb_first p2
				andalso c1 + 1 = c2
			fun f (p, (_,[])) = (p, [p])
			  | f (p, (prev,ps)) =
			  	if succeeding (prev, p) then (p, ps)
				else (p, p :: prev :: ps)
			val input = collapse hutree
			val (_, tmp) = foldl f (([],0),[]) input
			val result = rev (List.last input :: tmp) (* add last item *)
		in
			map stringify result
		end
		
	(* self test *)
	
	val tests = 
		let val tests_tree = node (leaf 10, node (node (leaf 20, node (leaf 40, none)), node (leaf 30, none)))
		in [
		insert none [true] 10 = node (none, leaf 10),
		insert none [false] 10 = node (leaf 10, none),
		insert none [true, true] 10 = node (none, node (none, leaf 10)),
		insert none [true, false] 10 = node (none, node (leaf 10, none)),
		insert (node (none, leaf 10)) [false] 20 = node (leaf 20, leaf 10),
		insert (node (none, leaf 10)) [false,true] 20 = node (node (none, leaf 20), leaf 10),
		code tests_tree [false] = SOME 10,
		code tests_tree [true, false, false] = SOME 20,
		code tests_tree [true, false, true, false] = SOME 40,
		code tests_tree [true, false, true, true] = NONE,
		code tests_tree [true, true, false] = SOME 30,
		code tests_tree [true, true, true] = NONE,
		has_step (node (none,none)),
		has_step (leaf 0) = false,
		has_step (none) = false,
		step (node (leaf 0, none)) false = leaf 0,
		step (node (none, leaf 1)) true = leaf 1,
		has_length tests_tree 1,
		has_length tests_tree 2,
		has_length tests_tree 3,
		has_length tests_tree 4,
		has_length tests_tree 5 = false,
		collapse none = [],
		collapse (node (leaf 1, none)) = [([false],1)],
		collapse (node (node (leaf 1,leaf 2), leaf 3)) = 
			[([false,false],1),([false,true],2),([true],3)]
		]
		end
		
end
